<?php
/**
 * SeekQuarry/Yioop --
 * Open Source Pure PHP Search Engine, Crawler, and Indexer
 *
 * Copyright (C) 2009 - 2018  Chris Pollett chris@pollett.org
 *
 * LICENSE:
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * END LICENSE
 *
 * @author Chris Pollett chris@pollett.org (initial MediaJob class
 *      and subclasses based on work of Pooja Mishra for her master's)
 * @license https://www.gnu.org/licenses/ GPL3
 * @link https://www.seekquarry.com/
 * @copyright 2009 - 2018
 * @filesource
 */
namespace seekquarry\yioop\library\media_jobs;

use seekquarry\yioop\configs as C;
use seekquarry\yioop\library as L;
use seekquarry\yioop\library\CrawlConstants;
use seekquarry\yioop\library\FetchUrl;
use seekquarry\yioop\library\IndexShard;
use seekquarry\yioop\library\PhraseParser;
use seekquarry\yioop\library\UrlParser;
use seekquarry\yioop\models\GroupModel;

/**
 * A media job to download and index feeds from various search sources (RSS,
 * HTML scraper, etc). Idea is that this job runs once an hour to get the
 * latest news, movies, weather from those sources.
 */
class WikiMediaJob extends MediaJob
{
    /**
     * how long in seconds before a feed item expires
     */
    const ITEM_EXPIRES_TIME = C\ONE_WEEK;
    /**
     * Mamimum number of feeds to download in one try
     */
    const MAX_PODCASTS_ONE_GO = 100;
    /**
     * Time in current epoch when feeds last updated
     * @var int
     */
    public $update_time;
    /**
     * Datasource object used to run db queries related to fes items
     * (for storing and updating them)
     * @var object
     */
    public $db;
    /**
     * Initializes the last update time to far in the past so, feeds will get
     * immediately updated. Sets up connect to DB to store feeds items, and
     * makes it so the same media job runs both on name server and client
     * Media Updaters
     */
    public function init()
    {
        $this->update_time = 0;
        $this->name_server_does_client_tasks = true;
        $this->name_server_does_client_tasks_only = true;
        $db_class = C\NS_DATASOURCES . ucfirst(C\DBMS). "Manager";
        $this->db = new $db_class();
        $this->db->connect();
        if(!empty($this->controller)) {
            $group_model = $this->controller->model("group");
        } else {
            $group_model = new GroupModel();
        }
        $this->group_model = $group_model;
        C\nsconddefine("WIKI_UPDATE_INTERVAL", C\ONE_DAY);
    }
    /**
     * Only update if its been more than an hour since the last update
     *
     * @return bool whether its been an hour since the last update
     */
    public function checkPrerequisites()
    {
        $time = time();
        $something_updated = false;
        $delta = $time - $this->update_time;
        if ($delta > C\WIKI_UPDATE_INTERVAL) {
            $this->update_time = $time;
            L\crawlLog("Performing medias podcasts update");
            return true;
        }
        L\crawlLog("Time since last update not exceeded, skipping wiki
            update");
        return false;
    }
    /**
     * Get the media sources from the local database and use those to run the
     * the same task as in the distributed setting
     */
    public function nondistributedTasks()
    {
        $db = $this->db;
        $sql = "SELECT * FROM MEDIA_SOURCE WHERE (TYPE='feed_podcast' OR ".
            "TYPE='scrape_podcast')";
        $result = $db->execute($sql);
        $podcasts = [];
        while ($podcast = $db->fetchArray($result)) {
            $this->parsePodcastAuxInfo($podcast);
            if (isset($podcast["PREVIOUSLY_DOWNLOADED"])) {
                $podcasts[] = $podcast;
            }
        }
        $this->tasks = $podcasts;
        $this->doTasks($podcasts);
    }
    /**
     * For each feed source downloads the feeds, checks which items are
     * not in the database, adds them. Then calls the method to rebuild the
     * inverted index shard for feeds
     *
     * @param array $tasks array of feed info (url to download, paths to
     *  extract etc)
     */
    public function doTasks($tasks)
    {
        if (!is_array($tasks)) {
            L\crawlLog(
                "----This media updater is NOT responsible for any podcasts!");
            return;
        }
        $podcasts = $tasks;
        L\crawlLog("----This media updater is responsible for the podcasts:");
        $i = 1;
        foreach ($podcasts as $podcast) {
            L\crawlLog("----  $i. " . $podcast["NAME"]);
            $i++;
        }
        $num_podcasts = count($podcasts);
        $podcasts_one_go = self::MAX_PODCASTS_ONE_GO;
        $limit = 0;
        while ($limit < $num_podcasts) {
            $podcasts_batch = array_slice($podcasts, $limit,
                $podcasts_one_go);
            $this->updatePodcastsOneGo($podcasts_batch);
            $limit += $podcasts_one_go;
        }
    }
    /**
     * @param array &$podcast
     */
    public function parsePodcastAuxInfo(&$podcast)
    {
        $locale_tag = $podcast["LANGUAGE"];
        $aux_parts = explode("###",
                html_entity_decode($podcast['AUX_INFO'], ENT_QUOTES));
        list($podcast['AUX_URL_XPATH'], , , , $podcast['LINK_XPATH'],
            $podcast['WIKI_PAGE'], $podcast['MAX_AGE']) = $aux_parts;
        $podcast['MAX_AGE'] = (empty($podcast['MAX_AGE'])) ?
            self::ITEM_EXPIRES_TIME : $podcast['MAX_AGE'];
        $group_model = $this->group_model;
        list($group_id, $page_id, $sub_path, $podcast["WIKI_FILE_PATTERN"]) =
            $group_model->getGroupIdPageIdSubPathFromName($podcast['WIKI_PAGE'],
            $locale_tag);
        $podcast["WIKI_PAGE_FOLDERS"] =
            $group_model->getGroupPageResourcesFolders($group_id, $page_id,
            $sub_path, true);
        if (empty($podcast["WIKI_PAGE_FOLDERS"][1])) {
            return;
        }
        $podcast["PREVIOUSLY_DOWNLOADED"] = [];
        list($resource_folder, $thumb_folder,) = $podcast["WIKI_PAGE_FOLDERS"];
        $podcast_download_file = $thumb_folder . "/" .
            L\crawlHash($podcast["NAME"]) . ".txt";
        if (file_exists($podcast_download_file)) {
            $podcast["PREVIOUSLY_DOWNLOADED"] = unserialize(file_get_contents(
                $podcast_download_file));
            $previously_downloaded = [];
            $max_age = $podcast["MAX_AGE"];
            $time = time();
            foreach ($podcast["PREVIOUSLY_DOWNLOADED"] as $guid => $item) {
                if ($max_age < 0||$item["SAVE_TIMESTAMP"] + $max_age > $time) {
                    $previously_downloaded[$guid] = $item;
                } else {
                    $group_model->deleteResource($item['FILENAME'], $group_id,
                        $page_id, $sub_path);
                }
            }
            $podcast["PREVIOUSLY_DOWNLOADED"] = $previously_downloaded;
        }
    }
    /**
     * @param array &$podcast
     * @param int $age
     */
    public function updatePodcastsOneGo($podcasts, $age = C\ONE_WEEK)
    {
        $podcasts = FetchUrl::getPages($podcasts, false, 0, null, "SOURCE_URL",
            CrawlConstants::PAGE, false, null, true);
        foreach ($podcasts as $podcast) {
            if (empty($podcast[CrawlConstants::PAGE])) {
                L\crawlLog("...No data in {$podcast['NAME']} feed skipping.");
                continue;
            }
            L\crawlLog("----Updating {$podcast['NAME']}.");
            // strip namespaces
            $page = preg_replace('@<(/?)(\w+\s*)\:@u', '<$1',
                $podcast[CrawlConstants::PAGE]);
            if (!empty($page)) {
                $podcast[CrawlConstants::PAGE] = $page;
            }
            $mime_type = (empty($podcast[CrawlConstants::TYPE])) ?
                "application/xml" : $podcast[CrawlConstants::TYPE];
            $is_html = preg_match("/html/i", $mime_type);
            if ($is_html && !empty($podcast['LINK_XPATH'])) {
                $num_added = $this->processHtmlPodcast($podcast, $age);
            } else if (!$is_html) {
                $num_added = $this->processFeedPodcast($podcast, $age);
            } else {
                $num_added = 0;
            }
            if ($num_added > 0) {
                $podcast_download_file = $podcast["WIKI_PAGE_FOLDERS"][1] .
                    "/" . L\crawlHash($podcast["NAME"]) . ".txt";
                file_put_contents($podcast_download_file,
                    serialize($podcast["PREVIOUSLY_DOWNLOADED"]));
            } else {
                L\crawlLog("----could not parse any podcasts from page.");
            }
        }
    }
    /**
     * @param array &$podcast
     * @param int $age
     */
    public function processHtmlPodcast(&$podcast, $age)
    {
        $page = $podcast[CrawlConstants::PAGE];
        $dom = $this->createDOMDocument($page);
        $source_url = $podcast["SOURCE_URL"];
        if (!empty($podcast['AUX_URL_XPATH'])) {
            $sub_aux_xpaths = explode("##", $podcast['AUX_URL_XPATH']);
            foreach ($sub_aux_xpaths as $aux_xpath) {
                $aux_url = $this->getLinkFromQueryPage($aux_xpath,
                    $page, $dom, $source_url);
                L\crawlLog("Downloading aux_url $aux_url");
                $page = FetchUrl::getPage($aux_url);
                $dom = $this->createDOMDocument($page);
            }
        }
        $url = $this->getLinkFromQueryPage($podcast['LINK_XPATH'], $page, $dom,
             $source_url);
        if (empty($aux_url)) {
            $aux_url = $url;
        }
        $item = ['guid' => L\crawlHash($url), 'link' => $url,
            'pubdate' => time(), 'title' => $aux_url];
        $did_add = $this->downloadPodcastItemIfNew($item, $podcast,
            $age);
        if ($did_add) {
            return 1;
        }
    }
    /**
     * @param string $page xml page to make DOM object for
     */
    function createDOMDocument($page)
    {
        $dom = new \DOMDocument();
        set_error_handler(null);
        @$dom->loadHTML('<?xml encoding="UTF-8">' . $page);
        set_error_handler(C\NS_CONFIGS . "yioop_error_handler");
        foreach ($dom->childNodes as $item) {
            if ($item->nodeType == XML_PI_NODE) {
                $dom->removeChild($item);
            }
        }
        return $dom;
    }
    /**
     *
     */
    public function getLinkFromQueryPage($xpath, $page, $dom, $source_url)
    {
        $dom_xpath = new \DOMXPath($dom);
        set_error_handler(null);
        $nodes = @$dom_xpath->evaluate($xpath);
        set_error_handler(C\NS_CONFIGS . "yioop_error_handler");
        $url = false;
        if ($nodes === false) {
            $regex_json_parts = explode("json|", $xpath);
            preg_match_all($regex_json_parts[0], $page, $matches);
            if (!empty($matches[1][0])) {
                $url = $matches[1][0];
                if (!empty($regex_json_parts[1])) {
                    $json_data = json_decode(trim($url, ";"), true);
                    $play_list_path = explode("|", $regex_json_parts[1]);
                    foreach ($play_list_path as $path) {
                        if (!empty($json_data[$path])) {
                            $json_data = $json_data[$path];
                        }
                    }
                    $url = $json_data;
                }
            }
        } else if ($nodes && $nodes->item(0)) {
            $node_item = $nodes->item(0);
            if (in_array($node_item->nodeName, ["a", "meta", "link"])) {
                $href = $node_item->attributes->getNamedItem("href");
                if(!empty($href)) {
                    $url = $href->nodeValue;
                }
                if (!$url && $node_item->nodeName == 'meta') {
                    $content = $node_item->attributes->getNamedItem("content");
                    if(!empty($content)) {
                        $url = $content->nodeValue;
                    }
                }
            } else if (in_array($node_item->nodeName, ["video", "audio"])) {
                $src = $node_item->attributes->getNamedItem("src");
                if(!empty($src)) {
                    $url = $src->nodeValue;
                }
            }
            if (!$url) {
                $url = $nodes->item(0)->nodeValue;
            }
        }
        if ($url) {
            $url = UrlParser::canonicalLink(
                urldecode(html_entity_decode($url)), $source_url);
        }
        return $url;
    }
    /**
     *
     * @param array &$podcast
     * @param int $age
     */
    public function processFeedPodcast(&$podcast, $age)
    {
        $page = $podcast[CrawlConstants::PAGE];
        $page = preg_replace("@<link@", "<slink", $page);
        $page = preg_replace("@</link@", "</slink", $page);
        $page = preg_replace("@pubDate@i", "pubdate", $page);
        $page = preg_replace("@&lt;@", "<", $page);
        $page = preg_replace("@&gt;@", ">", $page);
        $page = preg_replace("@<!\[CDATA\[(.+?)\]\]>@s", '$1', $page);
        // we also need a hack to make UTF-8 work correctly
        $dom = $this->createDOMDocument($page);
        $dom->encoding = 'UTF-8';
        $nodes = $dom->getElementsByTagName('item');
        // see above comment on why slink rather than link
        $item_elements = ["title" => "title",
            "description" => "description", "link" =>"slink",
            "guid" => "guid", "pubdate" => "pubdate"];
        if ($nodes->length == 0) {
            // maybe we're dealing with atom rather than rss
            $nodes = $dom->getElementsByTagName('entry');
            $item_elements = [
                "title" => "title",
                "description" => ["summary", "content"],
                "link" => "slink", "guid" => "id",
                "pubdate" => "updated"];
        }
        if (!empty($podcast['LINK_XPATH'])) {
            $item_elements['link'] = $podcast['LINK_XPATH'];
        }
        L\crawlLog("----...done extracting info. Check for new ".
            "podcast in {$podcast['NAME']}.");
        $num_added = 0;
        $num_seen = 0;
        foreach ($nodes as $node) {
            $item = [];
            foreach ($item_elements as $db_element => $podcast_element) {
                L\crawlTimeoutLog(
                    "----still adding podcast items to index.");
                if ($db_element == "link" && substr($podcast_element, 0, 4)
                    == "http") {
                    $element_text = $podcast_element;
                    $item[$db_element] = strip_tags($element_text);
                    continue;
                }
                if (!is_array($podcast_element)) {
                    $podcast_element = [$podcast_element];
                }
                foreach ($podcast_element as $tag_name) {
                    $tag_node = $node->getElementsByTagName(
                            $tag_name)->item(0);
                    $element_text = (is_object($tag_node)) ?
                        $tag_node->nodeValue: "";
                    if ($element_text) {
                        break;
                    }
                }
                if ($db_element == "link" && $tag_node &&
                    empty($element_text)) {
                    $element_text = $tag_node->getAttribute("href");
                    if (empty($element_text)) {
                        $element_text = $tag_node->getAttribute("url");
                    }
                    if (empty($element_text)) {
                        $element_text = $tag_node->getAttribute("src");
                    }
                    $element_text = UrlParser::canonicalLink($element_text,
                        $podcast["SOURCE_URL"]);
                }
                $item[$db_element] = strip_tags($element_text);
            }
            if (!empty($item['link']) && !empty($item['title']) &&
                !empty($item_elements['link'])
                && $item['link'] == $item_elements['link']) {
                $item['link'] .= "#" . L\crawlHash($item['title']);
            }
            if (empty($item['guid'])) {
                $item['guid'] = L\crawlHash($item['link'] . $item['title']);
            }
            $did_add = $this->downloadPodcastItemIfNew($item, $podcast, $age);
            if ($did_add) {
                $num_added++;
            }
            $num_seen++;
        }
        return $num_added;
    }
    /**
     *
     * @param array $item
     * @param array &$podcast
     * @param int $age
     */
    public function downloadPodcastItemIfNew($item, &$podcast, $age)
    {
        $group_model = $this->group_model;
        $pubdate = (empty($item['pubdate'])) ? time():
            (is_int($item['pubdate']) ? $item['pubdate'] :
            strtotime($item['pubdate']));
        if ($pubdate + $age < time()) {
            L\crawlLog("Podcast Item: {$item['title']} is too old skipping...");
            return false;
        }
        if (key_exists($item['guid'], $podcast["PREVIOUSLY_DOWNLOADED"])) {
            L\crawlLog("Podcast Item: {$item['title']} already downloaded, ".
                "skipping...");
            return false;
        }
        $type = UrlParser::getDocumentType($item['link']);
        $file_name = UrlParser::getDocumentFilename($item['link']);
        if (empty($file_name)) {
            $file_name = date("Y-m-d-H-i");
        }
        if (!empty($type)) {
            $file_name .= ".$type";
        }
        list($group_id, $page_id, $sub_path, $file_pattern) =
            $group_model->getGroupIdPageIdSubPathFromName($podcast['WIKI_PAGE'],
            $podcast['LANGUAGE']);
        $file_name = $this->makeFileNamePattern($file_name, $file_pattern,
            substr($item['title'], 0, C\NAME_LEN), $pubdate);
        $type = UrlParser::getDocumentType($file_name);
        $data = $this->downloadPodcastItem($item['link'], $type);
        if ($data) {
            $group_model->copyFileToGroupPageResource("", $file_name,
                L\mimeType($file_name), $group_id, $page_id, $sub_path, $data);
            $podcast["PREVIOUSLY_DOWNLOADED"][$item['guid']] = [
                "FILENAME" => $file_name,
                "SAVE_TIMESTAMP" => time()
            ];
            return true;
        }
        return false;
    }
    /**
     *
     * @param string $url
     * @param string $type
     */
    public function downloadPodcastItem($url, $type = "mp4")
    {
        L\crawlLog("Downloading ...$url");
        $data = $this->getPage($url);
        if (substr($data, 0, 6) != "#EXTM3") {
            return $data;
        }
        $convert_folder = C\WORK_DIRECTORY . self::CONVERT_FOLDER;
        if (!$this->makeFolder($convert_folder)) {
            return;
        }
        $convert_folder .= "/" . L\crawlHash($url);
        if (!$this->makeFolder($convert_folder)) {
            return;
        }
        if (stripos($data, "#EXT-X-STREAM-INF") !== false) {
            $lines = preg_split('/(\r)?\n/', $data);
            $max_url = "";
            $max_res = 0;
            $num_lines = count($lines);
            for ($i = 0; $i  < $num_lines; $i++) {
                if (preg_match('/\s*\#EXT\-X\-STREAM\-INF(.*)/i', $lines[$i],
                    $matches)) {
                    if (empty($max_url) && !empty($lines[$i + 1])) {
                        $max_url = $lines[$i + 1];
                    }
                }
                if (!empty($matches[1]) &&
                    preg_match('/RESOLUTION\=(.+)x(.+)/', $matches[1],
                    $sub_matches)) {
                    $width = intval($sub_matches[1]);
                    $height = intval($sub_matches[2]);
                    $res = $width * $height;
                    if ($max_res < $res && !empty($lines[$i + 1])) {
                        $max_url = $lines[$i + 1];
                        $max_res = $res;
                    }
                }
            }
            if (empty($max_url)) {
                return;
            }
            $max_url = UrlParser::canonicalLink($max_url, $url);
            return $this->downloadPodcastItem($max_url, $type);
        }
        $segments_txt_file = $convert_folder . "/segments.txt";
        file_put_contents($segments_txt_file, "");
        // find the video segments file for today's news
        $seen = [];
        $playlists = preg_split("/\#EXT\-X\-MEDIA\-SEQUENCE\s*\:\s*\d*/",
            $data);
        if (count($playlists) > 0) {
            array_shift($playlists);
        }
        $playlist = $playlists[0];
        //get any encryption file if needed
        preg_match_all('/\#EXT\-X\-KEY\:METHOD\=AES\-128,URI\=\"(.*)\"/',
            $playlist, $key_file_matches);
        if (!empty($key_file_matches[1][0])) {
            $key_file = $key_file_matches[1][0];
            $key_url = UrlParser::canonicalLink($key_file, $url);
            L\crawlLog("Downloading encrypt key url... $key_url");
            $key = FetchUrl::getPage($key_url);
        }
        // Now get playlist urls to download
        $playlist_urls = [];
        $lines = preg_split('/(\r)?\n/', $playlist);
        $num_lines = count($lines);
        for ($i = 0; $i  < $num_lines; $i++) {
            if (stristr($lines[$i], "#EXTINF") && !empty($lines[$i + 1])) {
                $playlist_urls[] = trim($lines[$i + 1]);
            }
        }
        foreach ($playlist_urls as $playlist_url) {
            $file_name = UrlParser::getDocumentFilename($playlist_url) . "." .
                UrlParser::getDocumentType($playlist_url);
            if (!file_exists("$convert_folder/$file_name")) {
                $item_url = UrlParser::canonicalLink($playlist_url, $url);
                L\crawlLog("Downloading item url... $item_url");
                $data = $this->getPage($item_url);
                if (!empty($key)) {
                    $data = openssl_decrypt($data, "AES-128-CBC", $key,
                        OPENSSL_RAW_DATA);
                }
                file_put_contents("$convert_folder/$file_name", $data);
                chmod("$convert_folder/$file_name", 0777);
            }
            file_put_contents($segments_txt_file,
                "file '$convert_folder/$file_name'\n", FILE_APPEND);
        }
        $out_file = "$convert_folder/out.$type";
        $ffmpeg = "ffmpeg -f concat -safe 0 -i $segments_txt_file " .
            " -c copy $out_file";
        L\crawlLog($ffmpeg);
        exec($ffmpeg);
        chmod($out_file, 0777);
        $data = file_get_contents($out_file);
        $model = new GroupModel();
        $model->db->unlinkRecursive($convert_folder);
        return $data;
    }
    /**
     *
     * @param string $file_name
     * @param string $file_pattern
     * @param int $time
     * @param int $time
     */
    private function makeFileNamePattern($file_name, $file_pattern,
        $title = "", $pubdate = null)
    {
        $file_name_parts = explode("?", $file_name);
        $file_name = basename($file_name_parts[0]);
        $file_name = str_replace("'", "", $file_name);
        $file_name = str_replace("\"", "", $file_name);
        $file_name = str_replace("/", "-", $file_name);
        $title = str_replace("'", "", $title);
        $title = str_replace("\"", "", $title);
        $title = str_replace("/", "-", $title);
        if (empty($file_pattern)) {
            return $file_name;
        }
        if (!$pubdate) {
            $pubdate = time();
        }
        $pattern_parts = preg_split('/(\%\w)/u', $file_pattern, -1,
            PREG_SPLIT_DELIM_CAPTURE);
        $out_name = "";
        foreach ($pattern_parts as $pattern_part) {
            if (!empty($pattern_part[0]) && $pattern_part[0] == '%') {
                if ($pattern_part[1] == 'F') {
                    $out_name .= $file_name;
                } else if ($pattern_part[1] == 'T') {
                    $out_name .= $title;
                } else {
                    $out_name .= date($pattern_part[1], $pubdate);
                }
            } else {
                $out_name .= $pattern_part;
            }
        }
        //just in case the name came from a url with extra garbbage
        return $out_name;
    }
    /**
     * Makes a directory in a way compatible with yioop's error handling.
     *
     * @param string $folders name of directory/folder to create.
     * @return boolean whether directory was created
     */
    private function makeFolder($folder)
    {
        if (!file_exists($folder)) {
            set_error_handler(null);
            @mkdir($folder);
            set_error_handler(C\NS_CONFIGS . "yioop_error_handler");
            if (!file_exists($folder)) {
                L\crawlLog("----Unable to create folder. Bailing!");
                return false;
            }
        }
        return true;
    }
    /**
     *
     */
    private function getPage($url)
    {
        return FetchUrl::getPage($url, null, false, null,
            4 * C\SINGLE_PAGE_TIMEOUT);
    }
}
